import platform


class GamepadReader:
    def __init__(self):
        self.__operating_system = platform.system()
        if self.__operating_system == "Windows":
            from mistercar_input_devices.backend.windows.adapters.input_readers.gamepad_reader_adapter import GamepadReaderAdapter
            self.__reader = GamepadReaderAdapter()
        elif self.__operating_system == "Linux":
            from mistercar_input_devices.backend.linux.adapters.input_readers.gamepad_reader_adapter import GamepadReaderAdapter
            self.__reader = GamepadReaderAdapter()
        elif self.__operating_system == "Darwin":
            from mistercar_input_devices.backend.macos.adapters.input_readers.gamepad_reader_adapter import GamepadReaderAdapter
            self.__reader = GamepadReaderAdapter()

    def get_control_state(self, control):
        return self.__reader.get_control_state(control)

    def get_states_of_multiple_controls(self, controls_to_check):
        return self.__reader.get_states_of_multiple_controls(controls_to_check)



import ctypes
import threading
from ctypes import wintypes
from win32con import WM_QUIT, SM_XVIRTUALSCREEN, SM_YVIRTUALSCREEN, SM_CXVIRTUALSCREEN, SM_CYVIRTUALSCREEN, SM_CXSCREEN, \
    SM_CYSCREEN
from mistercar_input_devices.backend.windows.platform_specific.structures import RAWINPUTDEVICE, RAWINPUTHEADER, \
    RAWINPUT
from mistercar_input_devices.backend.windows.platform_specific.window_utils import register_window_class, \
    create_message_only_window, unregister_window_class

user32 = ctypes.windll.user32
kernel32 = ctypes.windll.kernel32

# Constants from the Windows API
RID_INPUT = 0x10000003
WM_INPUT = 0x00FF
RIDEV_INPUTSINK = 0x00000100

# Constants for raw mouse input

WHEEL_DELTA = 120

# usFlags

MOUSE_MOVE_RELATIVE = 0x00  # Mouse movement data is relative to the last mouse position.
# For further information about mouse motion, see the following Remarks section.

MOUSE_MOVE_ABSOLUTE = 0x01  # Mouse movement data is based on absolute position.
# For further information about mouse motion, see the following Remarks section.

MOUSE_VIRTUAL_DESKTOP = 0x02  # Mouse coordinates are mapped to the virtual desktop (for a multiple monitor system).
# For further information about mouse motion, see the following Remarks section.

MOUSE_ATTRIBUTES_CHANGED = 0x04  # Mouse attributes changed; application needs to query the mouse attributes.

MOUSE_MOVE_NOCOALESCE = 0x08  # This mouse movement event was not coalesced. Mouse movement events can be coalesced by default.
# Windows XP/2000: This value is not supported.

# usButtonFlags

# Left button changed to down.
RI_MOUSE_BUTTON_1_DOWN = 0x0001
RI_MOUSE_LEFT_BUTTON_DOWN = RI_MOUSE_BUTTON_1_DOWN

# Left button changed to up.
RI_MOUSE_BUTTON_1_UP = 0x0002
RI_MOUSE_LEFT_BUTTON_UP = RI_MOUSE_BUTTON_1_UP

# Right button changed to down.
RI_MOUSE_BUTTON_2_DOWN = 0x0004
RI_MOUSE_RIGHT_BUTTON_DOWN = RI_MOUSE_BUTTON_2_DOWN

# Right button changed to up.
RI_MOUSE_BUTTON_2_UP = 0x0008
RI_MOUSE_RIGHT_BUTTON_UP = RI_MOUSE_BUTTON_2_UP

# Middle button changed to down.
RI_MOUSE_BUTTON_3_DOWN = 0x0010
RI_MOUSE_MIDDLE_BUTTON_DOWN = RI_MOUSE_BUTTON_3_DOWN

# Middle button changed to up.
RI_MOUSE_BUTTON_3_UP = 0x0020
RI_MOUSE_MIDDLE_BUTTON_UP = RI_MOUSE_BUTTON_3_UP

# XBUTTON1 changed to down.
RI_MOUSE_BUTTON_4_DOWN = 0x0040

# XBUTTON1 changed to up.
RI_MOUSE_BUTTON_4_UP = 0x0080

# XBUTTON2 changed to down.
RI_MOUSE_BUTTON_5_DOWN = 0x0100

# XBUTTON2 changed to up.
RI_MOUSE_BUTTON_5_UP = 0x0200

# Raw input comes from a mouse wheel.
RI_MOUSE_WHEEL = 0x0400

# Raw input comes from a horizontal mouse wheel.
RI_MOUSE_HWHEEL = 0x0800


class RawMouseReader:
    def __init__(self):
        self._thread = None
        self._is_alive = True
        self._hwnd = None
        self._wndclass = None
        self._absolute_x = 0
        self._absolute_y = 0
        self._relative_x = 0
        self._relative_y = 0
        self._wheel_up = 0
        self._wheel_down = 0
        self._hwheel_left = 0
        self._hwheel_right = 0
        self._lock = threading.Lock()

    def start(self):
        self._thread = threading.Thread(target=self._run, args=())
        self._thread.start()

    def stop(self):
        self._is_alive = False
        user32.PostThreadMessageW(self._thread.ident, WM_QUIT, 0, 0)
        if self._thread.is_alive():
            self._thread.join()
        if self._hwnd is not None:
            user32.DestroyWindow(self._hwnd)
            unregister_window_class(self._wndclass)

    def get_absolute_position(self):
        with self._lock:
            return self._absolute_x, self._absolute_y

    def get_relative_movement(self):
        with self._lock:
            relative_x, relative_y = self._relative_x, self._relative_y
            self._relative_x = self._relative_y = 0
            return relative_x, relative_y

    def get_wheel_delta(self):
        with self._lock:
            wheel_up, wheel_down = self._wheel_up, self._wheel_down
            self._wheel_up = self._wheel_down = 0
            return wheel_up, wheel_down

    def get_hwheel_delta(self):
        with self._lock:
            hwheel_left, hwheel_right = self._hwheel_left, self._hwheel_right
            self._hwheel_left = self._hwheel_right = 0
            return hwheel_left, hwheel_right

    def _run(self):
        # Register window class and create message-only window
        self._wndclass = register_window_class(self._wndproc, "MyMessageOnlyWindowClass")
        self._hwnd = create_message_only_window(self._wndclass)

        # Define the mouse usage page and usage
        RID_USAGE_PAGE_GENERIC = 0x01
        RID_USAGE_MOUSE = 0x02

        # Register the raw input device
        rid = RAWINPUTDEVICE()
        rid.usUsagePage = RID_USAGE_PAGE_GENERIC
        rid.usUsage = RID_USAGE_MOUSE
        rid.dwFlags = RIDEV_INPUTSINK
        rid.hwndTarget = self._hwnd

        if user32.RegisterRawInputDevices(ctypes.byref(rid), 1, ctypes.sizeof(rid)) == False:
            raise ctypes.WinError(ctypes.get_last_error())

        # Message loop
        msg = wintypes.MSG()
        while self._is_alive and user32.GetMessageW(ctypes.byref(msg), None, 0, 0) != 0:
            user32.TranslateMessage(ctypes.byref(msg))
            user32.DispatchMessageW(ctypes.byref(msg))

    def _wndproc(self, hwnd, msg, wparam, lparam):
        if msg == WM_INPUT:
            raw = RAWINPUT()
            sz = ctypes.c_uint(ctypes.sizeof(raw))
            szHeader = ctypes.c_uint(ctypes.sizeof(RAWINPUTHEADER))
            result = user32.GetRawInputData(lparam, RID_INPUT, ctypes.byref(raw), ctypes.byref(sz), szHeader)
            if result == sz.value:
                with self._lock:
                    if raw.data.mouse.usFlags & MOUSE_MOVE_ABSOLUTE:
                        if raw.data.mouse.usFlags & MOUSE_VIRTUAL_DESKTOP:
                            left = user32.GetSystemMetrics(SM_XVIRTUALSCREEN)
                            top = user32.GetSystemMetrics(SM_YVIRTUALSCREEN)
                            width = user32.GetSystemMetrics(SM_CXVIRTUALSCREEN)
                            height = user32.GetSystemMetrics(SM_CYVIRTUALSCREEN)
                        else:
                            left = 0
                            top = 0
                            width = user32.GetSystemMetrics(SM_CXSCREEN)
                            height = user32.GetSystemMetrics(SM_CYSCREEN)

                        self._absolute_x = int((raw.data.mouse.lLastX / 65535.0) * width) + left
                        self._absolute_y = int((raw.data.mouse.lLastY / 65535.0) * height) + top
                    else:
                        self._relative_x += raw.data.mouse.lLastX
                        self._relative_y += raw.data.mouse.lLastY

                    if raw.data.mouse.union.Buttons.usButtonFlags & RI_MOUSE_WHEEL:
                        wheel_delta = self._normalize_wheel_data(raw.data.mouse.union.Buttons.usButtonData)
                        if wheel_delta > 0:
                            self._wheel_up += wheel_delta
                        else:
                            self._wheel_down += abs(wheel_delta)
                    elif raw.data.mouse.union.Buttons.usButtonFlags & RI_MOUSE_HWHEEL:
                        hwheel_delta = self._normalize_wheel_data(raw.data.mouse.union.Buttons.usButtonData)
                        if hwheel_delta > 0:
                            self._hwheel_right += hwheel_delta
                        else:
                            self._hwheel_left += abs(hwheel_delta)
            else:
                print("GetRawInputData didn't return correct size")

        wparam = wintypes.WPARAM(wparam)
        lparam = wintypes.LPARAM(lparam)
        return user32.DefWindowProcW(hwnd, msg, wparam, lparam)

    def _normalize_wheel_data(self, wheel_data):
        """
        The mouse wheel delta values come in as a multiple of WHEEL_DELTA (120).
        If the scroll is upwards or to the right, the value is positive,
        while downwards or to the left is negative.
        However, these values are unsigned 16-bit integers. In the case of
        downward/leftward scroll, Windows represents it as a two's complement
        negative number.
        This helper function takes the raw wheel data, figures out if it's really
        a negative number in disguise, and if so converts it to standard negative
        format. Finally, it normalizes the delta to -1 (for down/left) or +1
        (for up/right).
        """
        if wheel_data > 32767:
            wheel_data -= 65536
        return wheel_data // WHEEL_DELTA



from typing import Optional
from mistercar_input_devices.backend.windows.platform_specific.keys_and_mouse import KeysAndMouse


class MouseEmulatorAdapter:
    def __init__(self):
        self._keys_and_mouse = KeysAndMouse()

    def move_mouse_by(self, dx: int, dy: int):
        self._keys_and_mouse.move_mouse_by(dx, dy)

    def move_cursor_to(self, x: int, y: int):
        self._keys_and_mouse.move_cursor_to(x, y)

    def scroll(self, clicks: int = 1):
        self._keys_and_mouse.scroll(clicks)

    def left_button_press(self, x: Optional[int] = None, y: Optional[int] = None):
        self._keys_and_mouse.left_button_press(x, y)

    def left_button_release(self, x: Optional[int] = None, y: Optional[int] = None):
        self._keys_and_mouse.left_button_release(x, y)

    def middle_button_press(self, x: Optional[int] = None, y: Optional[int] = None):
        self._keys_and_mouse.middle_button_press(x, y)

    def middle_button_release(self, x: Optional[int] = None, y: Optional[int] = None):
        self._keys_and_mouse.middle_button_release(x, y)

    def right_button_press(self, x: Optional[int] = None, y: Optional[int] = None):
        self._keys_and_mouse.right_button_press(x, y)

    def right_button_release(self, x: Optional[int] = None, y: Optional[int] = None):
        self._keys_and_mouse.right_button_release(x, y)

    def xbutton1_press(self, x: Optional[int] = None, y: Optional[int] = None):
        self._keys_and_mouse.xbutton1_press(x, y)

    def xbutton1_release(self, x: Optional[int] = None, y: Optional[int] = None):
        self._keys_and_mouse.xbutton1_release(x, y)

    def xbutton2_press(self, x: Optional[int] = None, y: Optional[int] = None):
        self._keys_and_mouse.xbutton2_press(x, y)

    def xbutton2_release(self, x: Optional[int] = None, y: Optional[int] = None):
        self._keys_and_mouse.xbutton2_release(x, y)

    def click(self, x: Optional[int] = None, y: Optional[int] = None, duration: float = 0.0):
        self._keys_and_mouse.click(x, y, duration)

    def left_click(self, x: Optional[int] = None, y: Optional[int] = None, duration: float = 0.0):
        self.click(x, y, duration)

    def middle_click(self, x: Optional[int] = None, y: Optional[int] = None, duration: float = 0.0):
        self._keys_and_mouse.middle_click(x, y, duration)

    def right_click(self, x: Optional[int] = None, y: Optional[int] = None, duration: float = 0.0):
        self._keys_and_mouse.right_click(x, y, duration)

    def xbutton1_click(self, x: Optional[int] = None, y: Optional[int] = None, duration: float = 0.0):
        self._keys_and_mouse.xbutton1_click(x, y, duration)

    def xbutton2_click(self, x: Optional[int] = None, y: Optional[int] = None, duration: float = 0.0):
        self._keys_and_mouse.xbutton2_click(x, y, duration)



import platform


class GamepadEmulator:
    def __init__(self):
        self.__operating_system = platform.system()
        if self.__operating_system == "Windows":
            from mistercar_input_devices.backend.windows.adapters.input_emulators.gamepad_emulator_adapter import GamepadEmulatorAdapter
            self.__emulator = GamepadEmulatorAdapter()
        elif self.__operating_system == "Linux":
            from mistercar_input_devices.backend.linux.adapters.input_emulators.gamepad_emulator_adapter import GamepadEmulatorAdapter
            self.__emulator = GamepadEmulatorAdapter()
        elif self.__operating_system == "Darwin":
            from mistercar_input_devices.backend.macos.adapters.input_emulators.gamepad_emulator_adapter import GamepadEmulatorAdapter
            self.__emulator = GamepadEmulatorAdapter()

    def emulate_control(self, control, value=None):
        """Set a value on the controller
        If percent is True all controls will accept a value between -1.0 and 1.0

        If not then:
            Triggers are 0 to 255
            Axis are -32768 to 32767

        Control List:
            AxisLx          , Left Stick X-Axis
            AxisLy          , Left Stick Y-Axis
            AxisRx          , Right Stick X-Axis
            AxisRy          , Right Stick Y-Axis
            BtnBack         , Menu/Back Button
            BtnStart        , Start Button
            BtnA            , A Button
            BtnB            , B Button
            BtnX            , X Button
            BtnY            , Y Button
            BtnThumbL       , Left Thumbstick Click
            BtnThumbR       , Right Thumbstick Click
            BtnShoulderL    , Left Shoulder Button
            BtnShoulderR    , Right Shoulder Button
            Dpad            , Set Dpad Value (0 = Off, Use DPAD_### Constants)
            TriggerL        , Left Trigger
            TriggerR        , Right Trigger
        """
        self.__emulator.emulate_control(control, value)

    def emulate_multiple_controls(self, control_value_dict):
        self.__emulator.emulate_multiple_controls(control_value_dict)



from ctypes import *
import os
import platform

if platform.architecture()[0] == '32bit':
    arc = '86'
else:
    arc = '64'

_path = os.path.abspath(os.path.join(
    os.path.dirname(__file__),
    'vXboxInterface-x{}'.format(arc),
    'vXboxInterface.dll'
))
_xinput = WinDLL(_path)


class MissingDependancyError(Exception):
    def __init__(self, message):
        self.message = message


if not _xinput.isVBusExists():
    raise MissingDependancyError(
        '''Unable to find VBus Controller.

Please refer to https://github.com/shauleiz/vXboxInterface/releases
or run "ScpVBus-x64/install.bat" in cmd.exe as administrator'''
    )

from .virtual_controller import vController
from .read_state import rController
from .virtual_controller import main as test_virtual
from .read_state import main as test_read



from typing import List, Optional
from mistercar_input_devices.device_analysis.analyzers.base import DeviceAnalyzer, DeviceInfo


class IOKitAnalyzer(DeviceAnalyzer):
    """macOS device analyzer using IOKit.

    Note:
        This is a placeholder implementation. Actual implementation would require
        using PyObjC to interface with IOKit or a custom C extension.
    """

    def __init__(self):
        super().__init__()
        self._device = None
        raise NotImplementedError(
            "macOS analyzer requires PyObjC or custom extension.\n"
            "Installation instructions will be provided in future updates."
        )

    def list_devices(self) -> List[DeviceInfo]:
        """List all available HID devices.

        Returns:
            List[DeviceInfo]: List of available devices

        Note:
            Would use IOHIDManagerCreate and IOHIDManagerSetDeviceMatching
            to enumerate devices.
        """
        raise NotImplementedError()

    def open_device(self, vendor_id: int, product_id: int) -> bool:
        """Open a specific device for analysis.

        Args:
            vendor_id: USB vendor ID of the device
            product_id: USB product ID of the device

        Returns:
            bool: True if device was opened successfully

        Note:
            Would use IOHIDDeviceCreate and related functions to open
            and configure the device.
        """
        raise NotImplementedError()

    def close_device(self) -> None:
        """Close the currently opened device.

        Note:
            Would use IOHIDDeviceClose and cleanup any resources.
        """
        if self._device:
            # Cleanup would go here
            self._device = None

    def read_data(self) -> Optional[bytes]:
        """Read a single data packet from the device.

        Returns:
            Optional[bytes]: Raw input data or None if no data available

        Note:
            Would use IOHIDDeviceRegisterInputValueCallback for asynchronous
            input handling.
        """
        raise NotImplementedError()



from mistercar_input_devices.backend.windows.platform_specific.keyboard import keyboard_reader as wkr


class KeyboardReaderAdapter:
    def __init__(self):
        self.__kb_reader = wkr.KeyboardReader()

    def get_key_state(self, key):
        return self.__kb_reader.get_key_state(key)

    def get_states_of_multiple_keys(self, keys_to_check):
        return self.__kb_reader.get_states_of_multiple_keys(keys_to_check)





class GamepadEmulatorAdapter:
    def __init__(self):
        pass

    def emulate_control(self, control, value):
        pass

    def emulate_multiple_controls(self, control_value_dict):
        pass



import ctypes
from ctypes import wintypes

user32 = ctypes.windll.user32
kernel32 = ctypes.windll.kernel32


WNDPROCTYPE = ctypes.WINFUNCTYPE(ctypes.c_longlong, wintypes.HWND, wintypes.UINT, wintypes.WPARAM, wintypes.LPARAM)


class WNDCLASSEX(ctypes.Structure):
    _fields_ = (
        ("cbSize", wintypes.UINT),
        ("style", wintypes.UINT),
        ("lpfnWndProc", WNDPROCTYPE),
        ("cbClsExtra", ctypes.c_int),
        ("cbWndExtra", ctypes.c_int),
        ("hInstance", wintypes.HINSTANCE),
        ("hIcon", wintypes.HICON),
        ("hCursor", ctypes.c_void_p),
        ("hbrBackground", wintypes.HBRUSH),
        ("lpszMenuName", wintypes.LPCWSTR),
        ("lpszClassName", wintypes.LPCWSTR),
        ("hIconSm", wintypes.HICON),
    )


def register_window_class(wndproc, class_name):
    wndclass = WNDCLASSEX()
    wndclass.cbSize = ctypes.sizeof(WNDCLASSEX)
    wndclass.lpfnWndProc = WNDPROCTYPE(wndproc)
    wndclass.lpszClassName = class_name
    wndclass.hInstance = kernel32.GetModuleHandleW(None)

    if not user32.RegisterClassExW(ctypes.byref(wndclass)):
        raise ctypes.WinError()

    return wndclass


# https://learn.microsoft.com/en-us/windows/win32/winmsg/window-features#message-only-windows
def create_message_only_window(wndclass):
    hwnd = user32.CreateWindowExW(
        0,  # dwExStyle
        wndclass.lpszClassName,  # lpClassName
        "Message-only window",  # lpWindowName
        0,  # dwStyle
        0,  # X
        0,  # Y
        0,  # nWidth
        0,  # nHeight
        ctypes.cast(-3, wintypes.HWND),  # hWndParent (HWND_MESSAGE)
        None,  # hMenu
        ctypes.c_void_p(wndclass.hInstance),  # hInstance
        None  # lpParam
    )

    if not hwnd:
        raise ctypes.WinError()

    return hwnd


def unregister_window_class(wndclass):
    user32.UnregisterClassW(wndclass.lpszClassName, kernel32.GetModuleHandleW(None))



from mistercar_input_devices.backend.windows.platform_specific.pyxinput import vController


class GamepadEmulatorAdapter:
    def __init__(self):
        self._gamepad = vController()

    def emulate_control(self, control, value):
        self._gamepad.set_value(control, value)

    def emulate_multiple_controls(self, control_value_dict):
        for key in control_value_dict:
            self._gamepad.set_value(key, control_value_dict[key])





class GamepadReaderAdapter:
    def __init__(self):
        pass

    def get_control_state(self, control):
        pass

    def get_states_of_multiple_controls(self, controls_to_check):
        pass



from mistercar_input_devices.backend.windows.platform_specific.pyxinput import rController


class GamepadReaderAdapter:
    def __init__(self):
        self.__read_controller = rController(1)

    def get_control_state(self, control):
        return self.__read_controller.gamepad.__getitem__(control)

    def get_states_of_multiple_controls(self, controls_to_check):
        states = [0] * len(controls_to_check)
        for i in range(len(controls_to_check)):
            states[i] = self.__read_controller.gamepad.__getitem__(controls_to_check[i])
        return states
